@@ -1,6 +1,6 @@
 // ***************************************************************** -*- C++ -*-
 /*
- * Copyright (C) 2004-2011 Andreas Huggel <ahuggel@gmx.net>
+ * Copyright (C) 2004-2015 Andreas Huggel <ahuggel@gmx.net>
  *
  * This program is part of the Exiv2 distribution.
  *
@@ -20,22 +20,17 @@
  */
 /*
   File:      rafimage.cpp
-  Version:   $Rev: 2455 $
+  Version:   $Rev: 3777 $
   Author(s): Andreas Huggel (ahu) <ahuggel@gmx.net>
   History:   05-Feb-07, ahu: created
   Credits:   See header file
  */
 // *****************************************************************************
 #include "rcsid_int.hpp"
-EXIV2_RCSID("@(#) $Id: rafimage.cpp 2455 2011-02-13 14:39:15Z ahuggel $")
+EXIV2_RCSID("@(#) $Id: rafimage.cpp 3777 2015-05-02 11:55:40Z ahuggel $")
 
-// *****************************************************************************
 // included header files
-#ifdef _MSC_VER
-# include "exv_msvc.h"
-#else
-# include "exv_conf.h"
-#endif
+#include "config.h"
 
 #include "rafimage.hpp"
 #include "tiffimage.hpp"
@@ -55,7 +50,7 @@ EXIV2_RCSID("@(#) $Id: rafimage.cpp 2455 2011-02-13 14:39:15Z ahuggel $")
 namespace Exiv2 {
 
     RafImage::RafImage(BasicIo::AutoPtr io, bool /*create*/)
-        : Image(ImageType::raf, mdExif | mdIptc, io)
+        : Image(ImageType::raf, mdExif | mdIptc | mdXmp, io)
     {
     } // RafImage::RafImage
 
